#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# @Script  : rpf_Shuffle.py


from utils.ribo import ArgsParser
from utils.ribo.Shuffle import *


def main():
    ArgsParser.now_time()
    print('\nShuffle the RPFs data.', flush=True)
    print('\nStep1: Checking the input Arguments.', flush=True)
    args = ArgsParser.shuffle_args_parser()
    rpfs = Shuffle(args)

    print('\nStep2: Import the RPFs.', flush=True)
    rpfs.import_rpf()

    print('\nStep3: Shuffle the RPFs table.', flush=True)
    rpfs.shuffle_rpfs()

    print('\nStep4: Output the RPFs table.', flush=True)
    rpfs.output_rpfs()

    ArgsParser.now_time()
    print('\nAll done.', flush=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Project      : RiboParser
@Script       : EndSite.py
@Environment  : python 3.12
@Version      : 1.0
@Author       : Rensc
@Time         : 2025/06/09 10:56:32
@E-mail       : rensc0718@163.com
@License      : (C)Copyright 2023-2025, Ren Shuchao
'''


# import pandas as pd
# import polars as pl
# import numpy as np
# from collections import OrderedDict
# from Bio import SeqIO
# import argparse



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : detect_offset.py


from utils.ribo import ArgsParser
from utils.ribo.Offset import *


def main():
    ArgsParser.now_time()
    print('\nDetect the p-site offset.', flush=True)
    print('\nStep1: Checking the input Arguments.', flush=True)
    args = ArgsParser.offset_args_parser()
    offset_attr = Offset(args)

    print('\nStep2: Import the transcripts annotation.', flush=True)
    offset_attr.read_transcript()

    print('\nStep3: Import the bam file.', flush=True)
    offset_attr.get_mrna_reads()

    # if args.mode == 'SSCBM':
    print('\nStep4: Detect the SSCBM offset of sequence profile.', flush=True)
    offset_attr.get_tis_offset()
    offset_attr.adjust_tis_offset()
    offset_attr.write_tis_offset()

    offset_attr.draw_tis_heatmap()

    print('\nStep5: Detect the RSBM offset of sequence profile.', flush=True)
    offset_attr.get_frame_offset()
    offset_attr.format_frame_offset()
    offset_attr.adjust_frame_offset()
    offset_attr.write_frame_offset()

    offset_attr.draw_frame_heatmap()

    # elif args.mode == 'RSBM':
    #     print('Step4: Detect the RSBM offset of sequence profile.\n', flush=True)
    #     offset_attr.get_frame_offset()
    #     offset_attr.format_frame_offset()
    #     offset_attr.adjust_frame_offset()
    #     offset_attr.write_frame_offset()

    #     offset_attr.draw_frame_heatmap()

    ArgsParser.now_time()
    print('\nAll done.', flush=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Project      : riboParser
@Script       : Shift.py
@Environment  : python 3.8.5
@Version      : 1.0
@Author       : Rensc 
@Time         : 2025/02/18 14:58:41
@E-mail       : rensc0718@163.com
@License      : (C)Copyright 2023-2025, Ren Shuchao
'''


from . import RPFs
from collections import OrderedDict
import matplotlib
import matplotlib.pyplot as plt
import polars as pl
import pandas as pd
import numpy as np
import seaborn as sns
import math


class Shift(object):

    def __init__(self, args):
        # input and output
        self.transcript = args.transcript
        self.rpf = args.rpf
        self.output = args.output

        # filter the range around tis/tts
        self.rpf_num = args.min
        self.tis = args.tis
        self.tts = args.tts

        # for the high expression gene filter
        self.sample_num = 0
        self.sample_name = None
        self.raw_rpf = None
        self.high_gene = None

        self.period = args.period
        self.period_count = None
                
        self.gene_frame_shift = pd.DataFrame(columns=['Sample', 'Frame', 'Count'])

    def import_rpf(self):
        '''
        @Message  : function for rpf import.
        @Input    : self.ribo --> rpf density file
                    self.tis --> tis range
                    self.tts --> tts range
                    self.rpf_num --> rpf number filter
        @Return   : output --> rpf density dataframe
        '''

        # raw_rpf, sample_name, sample_num, merged_rpf, total_rpf_num, gene_rpf_sum, high_gene, high_rpf
        rpf_results = RPFs.import_rpf(rpf_file=self.rpf,
                                      sites='P',
                                      frame='all',
                                      sample_num=None,
                                      sample_name=None,
                                      tis = self.tis,
                                      tts = self.tts,
                                      gene=self.transcript,
                                      rpf_num=self.rpf_num)
        
        self.raw_rpf = rpf_results[0].to_pandas()
        self.sample_name = rpf_results[1]
        self.sample_num = rpf_results[2]
        # self.merged_rpf = rpf_results[3]
        # self.total_rpf_num = rpf_results[4]
        # self.gene_rpf_sum = rpf_results[5]
        self.high_gene = rpf_results[6]
        # self.high_rpf = rpf_results[7]

        del rpf_results

    def calc_3nt_period(self):
        '''
        @Message  : function for each gene 3nt periodicity calculation.
        @Input    : param --> rpf density file
        @Return   : output --> summary the 3nt periodicity of each gene
        @Flow     : step1 --> calculate the count of each frame of samples
                    step2 --> calculate the ratio of each frame of samples
        '''

        raw_rpf = self.raw_rpf.drop(columns=["now_nt", "from_tis", "from_tts", "region", "codon"])

        # calculate the count of each frame of samples
        self.period_count  = raw_rpf.groupby('name').apply(np.sum, axis=0).drop(columns=["name"])
        # self.period_count = self.period_count .reset_index()

    def output_meta(self):
        out_txt = self.output + "_gene_periodicity.txt"
        self.period_count.to_csv(out_txt, sep='\t', index=True)

    def filter_frame_shift(self):
        '''
        @Message  : filter the frame shifting genes.
        @Input    : param --> rpf density file
                          --> frame shifting threshold
                          --> sample names
        @Return   : output --> frame shifting genes
        @Flow     : step1 --> calculate the frame shifting ratio of each gene
                    step2 --> filter the frame shifting genes of each sample
        '''
        
        def filter_frame_ratio(row, period):
            # filter the frame_shifting, two conditions:
            # if the frame1 > frame 0 ,frame1 > frame2, frame1 > self.period, then the gene is frame1 shifting
            # if the frame2 > frame 0 ,frame2 > frame1, frame2 > self.period, then the gene is frame2 shifting
            if row[0] > row[1] and row[0] > row[2] and row[0] > period:
                return 'frame0'
            if row[1] > row[0] and row[1] > row[2] and row[1] > period:
                return 'frame1'
            elif row[2] > row[0] and row[2] > row[1] and row[2] > period:
                return 'frame2'
            else:
                return 'fuzzy'

        # calculate the ratio of each frame of samples
        for sp in self.sample_name:

            output = self.output + "_" + sp + "_gene_frame_shift.txt"
            sp_colname = [sp + '_f0', sp + '_f1', sp + '_f2']

            period_ratio = self.period_count[sp_colname].apply(lambda x: x / x.sum() * 100, axis=1)
            period_ratio.loc[:, 'frame'] = period_ratio.apply(filter_frame_ratio, period=self.period, axis=1)
            
            period_ratio.to_csv(output, sep='\t', index=False)

            gene_frame_count = period_ratio['frame'].value_counts().reindex(['frame0', 'frame1', 'frame2'], fill_value=0).reset_index()
            gene_frame_count.insert(0, 'Sample', sp)
            gene_frame_count.rename(columns={'frame': 'Frame', 'count': 'Count'}, inplace=True)

            self.gene_frame_shift = pd.concat([self.gene_frame_shift, gene_frame_count], axis=0)

    def output_frame_shift(self):
        out_txt = self.output + "_gene_frame_shift_count.txt"
        self.gene_frame_shift.to_csv(out_txt, sep='\t', index=False)

    def get_sub_plot_num(self):
        factor1 = int(self.sample_num ** 0.5)
        factor2 = int(self.sample_num ** 0.5)
        square = factor1 * factor2

        while not square >= self.sample_num:
            if factor2 > factor1:
                factor1 += 1
            else:
                factor2 += 1
            square = factor1 * factor2

        return factor1, factor2
    
    def draw_frame_shift_count(self):

        out_pdf = self.output + "_gene_frame_shift_count_plot.pdf"
        out_png = self.output + "_gene_frame_shift_count_plot.png"

        nrow, ncol = self.get_sub_plot_num()
        
        matplotlib.use('AGG')
        fig, axes = plt.subplots(nrows=nrow, ncols=ncol, figsize=(nrow * 4, ncol * 3), sharey=True)

        for (sample, group), ax in zip(self.gene_frame_shift.groupby('Sample'), axes.flatten()):
            bars = group.plot(x='Frame', y='Count', kind='bar', ax=ax, title=sample, legend=False)
            ax.set_ylabel('Count')
            ax.set_xlabel('')

            for bar in bars.patches:
                height = bar.get_height()
                ax.annotate(f'{height:.0f}', 
                            xy=(bar.get_x() + bar.get_width() / 2, height),
                            xytext=(0, 3),  # 3 points vertical offset
                            textcoords="offset points",
                            ha='center', va='bottom')
        
        plt.tight_layout()

        # plt.show()
        fig.savefig(fname=out_pdf)
        fig.savefig(fname=out_png, dpi=300)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : rpf_Shift.py


from utils.ribo import ArgsParser
from utils.ribo import Shift

def main():
    ArgsParser.now_time()
    print('\nDraw the frame shifting plot.', flush=True)
    print('\nStep1: Checking the input Arguments.', flush=True)
    args = ArgsParser.frame_shift_args_parser()

    print('\nStep2: Import the RPFs file.', flush=True)
    rpfs = Shift.Shift(args)
    rpfs.import_rpf()

    print('\nStep3: Calculate the 3nt periodicity.', flush=True)
    rpfs.calc_3nt_period()

    print('\nStep4: Ouput the 3nt periodicity.', flush=True)
    rpfs.output_meta()

    print('\nStep5: Filter frame shift.', flush=True)
    rpfs.filter_frame_shift()

    print('\nStep6: Output the frame shift.', flush=True)
    rpfs.output_frame_shift()

    print('\nStep7: Draw the frame shifting plot.', flush=True)
    rpfs.draw_frame_shift_count()

    print('\nAll done.\n', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : Percentage.py


from concurrent.futures import ThreadPoolExecutor
from scipy.stats import zscore
from math import ceil
from collections import OrderedDict
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import pandas as pd
from . import RPFs


class Percentage(object):

    def __init__(self, args):
        # input and output
        self.rpf = args.rpf
        self.gene = args.transcript
        self.gene_table = None
        self.output = args.output

        # filter the stable codon coverage range
        self.tis = None
        self.tts = None
        self.site = 'P'
        self.frame = args.frame
        self.rpf_num = args.min

        self.norm = args.normal
        self.sample_name = None

        # for the high expression gene filter
        self.raw_rpf = None
        # self.merged_rpf = None
        # self.high_gene = None
        self.high_rpf = None
        self.gene_num = None
        self.gene_rpf_sum = None
        self.total_rpf_num = None

        # gene coverage
        self.gene_coverage = None


    def read_rpf(self):
        # raw_rpf, sample_name, sample_num, merged_rpf, total_rpf_num, gene_rpf_sum, high_gene, high_rpf
        rpf_results = RPFs.import_rpf(rpf_file=self.rpf,
                                      sites=self.site,
                                      frame=self.frame,
                                      sample_num=None,
                                      sample_name=None,
                                      tis=self.tis,
                                      tts=self.tts,
                                      gene=self.gene,
                                      rpf_num=self.rpf_num)
        # self.raw_rpf = rpf_results[0]
        self.sample_name = rpf_results[1].copy()
        self.sample_num = rpf_results[2]
        # self.merged_rpf = rpf_results[3]
        self.total_rpf_num = rpf_results[4]
        # self.gene_rpf_sum = rpf_results[5]
        self.high_gene = rpf_results[6]
        self.gene_num = len(self.high_gene)
        self.high_rpf = rpf_results[7].copy()
        del rpf_results

        if self.norm:
            self.high_rpf[self.sample_name] = self.high_rpf[self.sample_name].astype(float)

            self.high_rpf[self.sample_name] = self.high_rpf[self.sample_name].div(self.total_rpf_num) * 1e6


    def import_gene(self):
        '''
        @Message  : import the gene file
        @Input    : self.gene --> gene file
        @Return   : self.gene --> gene dataframe
        @Flow     : step1 --> import the gene file
        '''

        self.gene_table = pd.read_csv(self.gene, sep='\t', index_col=False)
        self.gene_table = self.gene_table.loc[self.gene_table['transcript_id'].isin(self.high_gene), ]


    def calc_density_percent(self):
        '''
        @Message  : calculate the gene rpf density percentage of each gene
        @Input    : gene_rpf --> gene rpf dataframe
        @Return   : gene_rpf --> gene rpf density dataframe
        @Flow     : step1 --> calculate the gene rpf density
        '''
        # remove the UTR5 and UTR3 region
        self.high_rpf = self.high_rpf.loc[self.high_rpf['region'] == 'cds', ]
        
        # group the gene by the gene name and sum the value
        self.gene_rpf_sum = self.high_rpf.groupby('name', observed=False)[self.sample_name].sum()

        # convert the rpf value to 1, if the value > 0
        self.high_rpf[self.sample_name] = self.high_rpf[self.sample_name].map(lambda x: 1 if x > 0 else 0)

        # group by the gene and sum the value, then divide the gene length
        gene_p_site_num = self.high_rpf.groupby('name', observed=False)[self.sample_name].sum()
        gene_length = self.high_rpf.groupby('name', observed=False)['from_tis'].max()
        self.gene_coverage = gene_p_site_num.div(gene_length, axis=0) * 100


    def draw_rpf_histogram(self):
        '''
        @Message  : draw the rpf histogram of the gene
        @Input    : self.gene_coverage --> the gene coverage dataframe
        @Return   : None
        @Flow     : step1 --> draw the rpf histogram of the gene
        '''

        # draw the gene coverage histogram of the sample
        for sp in self.sample_name:
            print('Draw percentage histogram of {sp}.'.format(sp=sp), flush=True)

            out_pdf = self.output + "_" + sp + "_coverage_histogram.pdf"
            out_png = self.output + "_" + sp + "_coverage_histogram.png"

            matplotlib.use('AGG')

            fig, axes = plt.subplots(1, 2, figsize=(8, 3.5), gridspec_kw={'width_ratios': [1, 1]})

            # draw the gene coverage histogram of the sample
            sns.histplot(self.gene_coverage[sp], bins=100, shrink=0.9, 
                         kde=True, edgecolor='white', color='#2b79da', ax=axes[0])
            axes[0].set_ylabel('Gene count', fontsize=12)
            axes[0].set_xlabel("Percentage (%)", fontsize=12)
            axes[0].set_title(('Coverage histogram ({number})').format(number=self.gene_num), fontsize=14)

            sns.histplot(np.log2(self.gene_rpf_sum[sp] + 1), bins=100, shrink=0.9, 
                         kde=True, edgecolor='white', color='#2b79da', ax=axes[1])
            axes[1].set_ylabel('Gene count', fontsize=12)
            axes[1].set_xlabel("log2(Expr + 1)", fontsize=12)
            axes[1].set_title(('Abundance histogram ({number})').format(number=self.gene_num), fontsize=14)

            fig.tight_layout()
            # plt.show()
            fig.savefig(fname=out_pdf)
            fig.savefig(fname=out_png)
            plt.close()


    def draw_rpf_boxplot(self):
        '''
        @Message  : draw the boxplot of the percentage
        @Input    : self.gene_coverage --> the gene coverage dataframe
        @Return   : None
        @Flow     : step1 --> draw the rpf boxplot of the gene
        '''

        gene_coverage = pd.melt(self.gene_coverage.reset_index(), id_vars=['name'], var_name='Sample', value_name='Percentage')

        gene_rpf_sum_log2 = np.log2(self.gene_rpf_sum + 1)
        gene_expression = pd.melt(gene_rpf_sum_log2.reset_index(), id_vars=['name'], var_name='Sample', value_name='Expression')

        # draw the gene coverage boxplot of the sample
        print('Draw percentage boxplot of RPFs coverage.', flush=True)

        out_pdf = self.output + "_coverage_boxplot.pdf"
        out_png = self.output + "_coverage_boxplot.png"

        width = 0 if self.sample_num < 6 else (self.sample_num - 6) * 0.3

        matplotlib.use('AGG')

        fig, axes = plt.subplots(1, 2, figsize=(7 + width, 5), gridspec_kw={'width_ratios': [1, 1]})

        # draw the gene coverage boxplot of the sample
        sns.boxplot(data = gene_coverage, x = 'Sample', y='Percentage', 
                    fliersize = 1, linewidth = 0.5,
                    color='#2b79da', ax=axes[0])
        axes[0].set_ylabel('Percentage (%)', fontsize=12)
        axes[0].set_xticklabels(axes[0].get_xticklabels(), rotation=90, ha='right', va='top', fontsize=12, color='black')
        axes[0].set_xlabel("Sample", fontsize=12)
        axes[0].set_title(('Coverage boxplot ({number})').format(number=self.gene_num), fontsize=14)

        sns.boxplot(data = gene_expression, x = 'Sample', y='Expression', 
                    fliersize = 1, linewidth = 0.5,
                    color='#2b79da', ax=axes[1])
        axes[1].set_ylabel('Expression (log2)', fontsize=12)
        axes[1].set_xticklabels(axes[1].get_xticklabels(), rotation=90, ha='right', va='top', fontsize=12, color='black')
        axes[1].set_xlabel("Sample", fontsize=12)
        axes[1].set_title(('Abundance boxplot ({number})').format(number=self.gene_num), fontsize=14)

        fig.tight_layout()
        # plt.show()
        fig.savefig(fname=out_pdf)
        fig.savefig(fname=out_png)
        plt.close()


    def output_density_percent(self):
        '''
        @Message  : output the gene rpf density percentage
        @Input    : self.gene_coverage --> the gene coverage dataframe
        @Return   : None
        @Flow     : step1 --> output the gene rpf density percentage
        '''

        # output the gene coverage of the sample
        out_file = self.output + "_gene_coverage_percent.txt"
        self.gene_coverage.to_csv(out_file, sep='\t', index=True, header=True)
        


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : Quant.py


import matplotlib.pyplot as plt
import seaborn as sns

# import plotly.graph_objs as go
# import plotly.express as px
from scipy.spatial.distance import pdist
from scipy.cluster import hierarchy

from sklearn.decomposition import PCA

import numpy as np
import pandas as pd
from . import RPFs


class Quant(object):
    def __init__(self, args):
        self.rpf_file = args.rpf

        # self.anno = args.anno
        self.output_prefix = args.output

        # parameters for the trimmed CDS region
        self.rpf_num = 0
        self.site = 'P'
        self.tis = args.tis
        self.tts = args.tts
        self.frame = args.frame
        self.length = None

        # parameters for rpf file parser
        self.sample_name = None
        self.sample_num = None
        self.merged_rpf = None
        self.total_rpf = None
        self.col_index = None

        # parameters for different region
        self.utr5 = args.utr5
        self.utr3 = args.utr3

        self.utr5_rpf = None
        self.utr5_rpm = None
        self.utr5_tpm = None

        self.utr3_rpf = None
        self.utr3_rpm = None
        self.utr3_tpm = None

        self.cds_rpf = None
        self.cds_rpm = None
        self.cds_tpm = None

    def read_rpf(self):
        # raw_rpf, sample_name, sample_num, merged_rpf, total_rpf_num, gene_rpf_sum, high_gene, high_rpf
        rpf_results = RPFs.import_rpf(rpf_file=self.rpf_file,
                                      sites=self.site,
                                      frame=self.frame,
                                      sample_num=None,
                                      sample_name=None,
                                      tis=None,
                                      tts=None,
                                      gene=None,
                                      rpf_num=self.rpf_num)
        # self.raw_rpf = rpf_results[0]
        self.sample_name = rpf_results[1].copy()
        # self.sample_num = rpf_results[2]
        self.merged_rpf = rpf_results[3].copy()
        self.total_rpf = rpf_results[4]
        # self.gene_rpf_sum = rpf_results[5]
        # self.high_gene = rpf_results[6]
        # self.high_rpf = rpf_results[7].copy()
        del rpf_results

        self.sample_num = len(self.sample_name)
        self.col_index = ['name'] + self.sample_name
        self.length = self.merged_rpf.loc[:, ['name', 'region']].groupby("name").count() * 3

    def output_total_rpf(self):
        total_rpf_name = self.output_prefix + "_total.txt"
        self.total_rpf.index.name = 'sample'
        self.total_rpf.name = 'rpf_count'

        self.total_rpf.to_csv(total_rpf_name, sep='\t', index=True)

    def output_gene_rpf(self, rpf_count, rpf_rpm, rpf_rpkm, rpf_tpm, region):

        # output the rpf / rpm / tpm results
        rpf_count.astype(int)
        rpf_count.index.name = "name"
        rpf_count.to_csv(self.output_prefix + "_" + region + "_rpf_quant.txt", sep='\t', index=True)

        rpf_rpm = rpf_rpm.astype(float).round(2)
        rpf_rpm.index.name = "name"
        rpf_rpm.to_csv(self.output_prefix + "_" + region + "_rpm_quant.txt", sep='\t', index=True)

        rpf_rpkm = rpf_rpkm.astype(float).round(2)
        rpf_rpkm.index.name = "name"
        rpf_rpkm.to_csv(self.output_prefix + "_" + region + "_rpkm_quant.txt", sep='\t', index=True)

        rpf_tpm = rpf_tpm.astype(float).round(2)
        rpf_tpm.index.name = "name"
        rpf_tpm.to_csv(self.output_prefix + "_" + region + "_tpm_quant.txt", sep='\t', index=True)

    def quant_utr5_rpf(self, region='5utr'):

        print("Summary RPFs of 5'-UTR.", flush=True)
        # filter the gene region
        self.utr5_rpf = self.merged_rpf.loc[self.merged_rpf["region"] == region, self.col_index]
        self.utr5_rpf = self.utr5_rpf.groupby("name")[self.sample_name].sum()

        self.utr5_rpf.columns = [i + "_" + region + "_rpf" for i in self.sample_name]

        # calculate the gene rpm
        self.utr5_rpm = np.divide(self.utr5_rpf, np.asarray(self.total_rpf)) * 1e6
        self.utr5_rpm.columns = [i + "_" + region + "_rpm" for i in self.sample_name]

        # calculate the gene rpkm
        self.utr5_rpkm = (self.utr5_rpf * 1e9).div(self.length['region'], axis=0).div(np.asarray(self.total_rpf), axis=1)
        self.utr5_rpkm.columns = [i + "_" + region + "_rpkm" for i in self.sample_name]

        # calculate the gene tpm
        utr5_rpk = (self.utr5_rpf * 1000).div(self.length['region'], axis=0).dropna()
        utr5_rpk_sum = utr5_rpk.sum(axis=0)
        self.utr5_tpm = (utr5_rpk * 1000000).div(utr5_rpk_sum, axis=1)
        self.utr5_tpm.columns = [i + "_" + region + "_tpm" for i in self.sample_name]

        self.output_gene_rpf(self.utr5_rpf, self.utr5_rpm, self.utr5_rpkm, self.utr5_tpm, 'utr5')

    def quant_utr3_rpf(self, region='3utr'):

        print("Summary RPFs of 3'-UTR.", flush=True)
        # filter the gene region
        self.utr3_rpf = self.merged_rpf.loc[self.merged_rpf["region"] == region, self.col_index]
        self.utr3_rpf = self.utr3_rpf.groupby("name")[self.sample_name].sum()

        self.utr3_rpf.columns = [i + "_" + region + "_rpf" for i in self.sample_name]

        # calculate the gene rpm
        self.quant_rpm = np.divide(self.utr3_rpf, np.asarray(self.total_rpf)) * 1e6
        self.quant_rpm.columns = [i + "_" + region + "_rpm" for i in self.sample_name]

        # calculate the gene rpkm
        self.utr3_rpkm = (self.utr3_rpf * 1e9).div(self.length['region'], axis=0).div(np.asarray(self.total_rpf), axis=1)
        self.utr3_rpkm.columns = [i + "_" + region + "_rpkm" for i in self.sample_name]

        # calculate the gene tpm
        utr3_rpk = (self.utr3_rpf * 1000).div(self.length['region'], axis=0).dropna()
        utr3_rpk_sum = utr3_rpk.sum(axis=0)
        self.utr3_tpm = (utr3_rpk * 1000000).div(utr3_rpk_sum, axis=1)
        self.utr3_tpm.columns = [i + "_" + region + "_tpm" for i in self.sample_name]

        self.output_gene_rpf(self.utr3_rpf, self.utr3_rpm, self.utr3_rpkm, self.utr3_tpm, 'utr3')

    def quant_cds_rpf(self, region='cds'):

        print('Summary RPFs of CDS.', flush=True)
        # filter the gene region
        self.cds_rpf = self.merged_rpf[self.merged_rpf['from_tis'] >= self.tis]
        self.cds_rpf = self.cds_rpf[self.cds_rpf['from_tts'] < -self.tts]

        self.cds_rpf = self.cds_rpf.loc[self.cds_rpf["region"] == region, self.col_index]
        self.cds_rpf = self.cds_rpf.groupby("name")[self.sample_name].sum()

        self.cds_rpf.columns = [i + "_" + region + "_rpf" for i in self.sample_name]

        # calculate the gene rpm
        self.cds_rpm = np.divide(self.cds_rpf, np.asarray(self.total_rpf)) * 1e6
        self.cds_rpm.columns = [i + "_" + region + "_rpm" for i in self.sample_name]

        # calculate the gene rpkm
        # formula: RPKM = (10^9 * C) / (N * L)
        # C: number of reads mapped to a gene
        # N: total number of reads mapped to the reference
        # L: number of bases in a gene
        
        self.cds_rpkm = (self.cds_rpf * 1e9).div(self.length['region'], axis=0).div(np.asarray(self.total_rpf), axis=1)
        self.cds_rpkm.columns = [i + "_" + region + "_rpkm" for i in self.sample_name]

        # calculate the gene tpm
        cds_rpk = (self.cds_rpf * 1000).div(self.length['region'], axis=0).dropna()
        cds_rpk_sum = cds_rpk.sum(axis=0)
        self.cds_tpm = (cds_rpk * 1000000).div(cds_rpk_sum, axis=1)
        self.cds_tpm.columns = [i + "_" + region + "_tpm" for i in self.sample_name]

        self.output_gene_rpf(self.cds_rpf, self.cds_rpm, self.cds_rpkm, self.cds_tpm, 'cds')

    def quant_region(self):
        if self.utr3:
            self.quant_utr5_rpf(region='utr5')

        self.quant_cds_rpf(region='cds')

        if self.utr5:
            self.quant_utr3_rpf(region='utr3')

    def get_sub_plot_num(self):
        factor1 = int(self.sample_num ** 0.5)
        factor2 = int(self.sample_num ** 0.5)
        square = factor1 * factor2

        while not square >= self.sample_num:
            if factor2 > factor1:
                factor1 += 1
            else:
                factor2 += 1
            square = factor1 * factor2

        return factor1, factor2

    # def draw_rpf_barplot(self):
    #     """
    #     draw the rpf region of each samples with plotly

    #     1. summary the rpf num of each region [5'-utr, cds, 3'-utr]
    #     2. draw the stacked bar plot
    #     """
        
    #     out_pdf = self.output_prefix + "_cds_rpm_bar_plot.pdf"
    #     # out_png = self.output_prefix + "_rpf_barplot.png"

    #     rpf_sum = self.merged_rpf.groupby('region')[self.sample_name].apply(sum)
    #     rpf_sum_per = round(rpf_sum.div(rpf_sum.sum()) * 100, 2)

    #     rpf_sum_per_t = rpf_sum_per.reset_index().melt(id_vars="region",
    #                                                    var_name="samples",
    #                                                    value_name="proportion")

    #     fig = px.bar(rpf_sum_per_t,
    #                  x="samples",
    #                  y="proportion",
    #                  color="region",
    #                  orientation="v",
    #                  text_auto=True,
    #                  category_orders={"region": ["5utr", "cds", "3utr"]},
    #                  color_discrete_sequence=px.colors.qualitative.Pastel2)

    #     fig.update_layout(xaxis_title="proportion (%)",
    #                       yaxis_title="samples",
    #                       template="simple_white",
    #                       height=500, width=300 + len(self.sample_name) * 20)

    #     fig.write_image(out_pdf, engine="kaleido")
    #     # fig.write_image(out_png, engine="kaleido")

    def draw_rpf_barplot(self):
        """
        Draw the RPF region of each sample using matplotlib & seaborn

        1. Summarize the RPF number of each region [5'-UTR, CDS, 3'-UTR]
        2. Draw the stacked bar plot
        """

        out_pdf = self.output_prefix + "_cds_rpm_bar_plot.pdf"

        # Step 1: Summarize
        rpf_sum = self.merged_rpf.groupby('region')[self.sample_name].apply(sum)
        rpf_sum_per = round(rpf_sum.div(rpf_sum.sum()) * 100, 2)

        rpf_sum_per_t = rpf_sum_per.reset_index().melt(
            id_vars="region",
            var_name="samples",
            value_name="proportion"
        )

        # Step 2: Draw
        plt.figure(figsize=(max(6, len(self.sample_name) * 0.8), 5))
        sns.barplot(
            data=rpf_sum_per_t,
            x="samples",
            y="proportion",
            hue="region",
            palette="Pastel2"
        )

        plt.xlabel("Samples")
        plt.ylabel("Proportion (%)")
        plt.xticks(rotation=45, ha="right")
        plt.legend(title="Region", bbox_to_anchor=(1.05, 1), loc='upper left')

        plt.tight_layout()
        plt.savefig(out_pdf)
        plt.close()
        

    def draw_rpf_cdfplot(self):
        """
        Draw the CDF figure of each sample using matplotlib & seaborn (no statsmodels)
        
        1. Log-transform the CDS RPM values
        2. Draw the eCDF plot for each sample
        """

        out_pdf = self.output_prefix + "_cds_rpm_cdf_plot.pdf"

        # Step 1: log2 transform
        cds_log = np.log2(self.cds_rpm + 1)
        cds_log = cds_log.reset_index().melt(
            id_vars="name",
            var_name="samples",
            value_name="rpm"
        )

        # Step 2: Draw eCDF for each sample
        plt.figure(figsize=(8, 6))
        palette = sns.color_palette("tab10", n_colors=len(cds_log["samples"].unique()))
        
        for color, sample in zip(palette, cds_log["samples"].unique()):
            values = np.sort(cds_log.loc[cds_log["samples"] == sample, "rpm"].values)
            y = np.arange(1, len(values) + 1) / len(values)
            plt.plot(values, y, label=sample, linewidth=1.5, color=color)

        plt.xlabel("log2 RPM")
        plt.ylabel("eCDF")
        plt.legend(title="Samples", bbox_to_anchor=(1.05, 1), loc='upper left')
        plt.tight_layout()
        plt.savefig(out_pdf)
        plt.close()

    # def draw_rpf_cdfplot(self):
    #     """
    #     draw the cdf figure of each samples with plotly
    #     to check the global expression levels

    #     1. log the cds rpm
    #     2. draw the eCDF plot of each sample with plotly
    #     """
    #     out_pdf = self.output_prefix + "_cds_rpm_cdf_plot.pdf"
    #     # out_png = self.output_prefix + "_rpf_cdfplot.png"

    #     # get the log2 values
    #     cds_log = np.log2(self.cds_rpm + 1)
    #     cds_log = cds_log.reset_index().melt(id_vars="name", var_name="samples", value_name="rpm")

    #     # draw the proportion of RPFs in different region
    #     fig = px.ecdf(cds_log, x="rpm", color="samples")

    #     fig.update_layout(xaxis_title="eCDF",
    #                       yaxis_title="log2 RPM",
    #                       template="simple_white",
    #                       height=600, width=650)

    #     fig.write_image(out_pdf, engine="kaleido")
    #     # fig.write_image(out_png, engine="kaleido")


    def draw_rpf_pcaplot(self):
        """
        Draw the PCA figure of each sample using matplotlib & seaborn

        1. Log-transform the CDS RPM
        2. Calculate PCA components
        3. Draw PCA scatter plot
        """

        out_pdf = self.output_prefix + "_cds_rpm_pca_plot.pdf"
        out_txt = self.output_prefix + "_cds_rpm_pca.txt"

        # Step 1: PCA calculation
        data_log = np.log2(self.cds_rpm.T + 1)  # transpose so samples are rows
        pca = PCA(n_components=2)
        components = pca.fit_transform(data_log)
        variance_ratios = pca.explained_variance_ratio_ * 100  # percentage

        # Step 2: Save PCA result
        pca_df = pd.DataFrame(data=components, columns=['PC1', 'PC2'])
        pca_df.index = self.sample_name
        pca_df = pca_df.reset_index().rename(columns={'index': 'samples'})
        pca_df.to_csv(out_txt, sep='\t', index=False)

        # Step 3: Draw scatter plot
        plt.figure(figsize=(6, 5))
        sns.scatterplot(
            data=pca_df,
            x="PC1", y="PC2",
            hue="samples",
            palette="Set2",
            s=80,
            edgecolor="black"
        )

        plt.xlabel(f"PC1 ({variance_ratios[0]:.2f}%)")
        plt.ylabel(f"PC2 ({variance_ratios[1]:.2f}%)")
        plt.legend(title="Samples", bbox_to_anchor=(1.05, 1), loc='upper left')
        plt.tight_layout()
        plt.savefig(out_pdf)
        plt.close()


    # def draw_rpf_pcaplot(self):
    #     """
    #     draw the pca figure of each samples with plotly

    #     1. log the cds rpm
    #     2. calculate the components of each dimension
    #     3. draw the PCA scatter plot of each sample
    #     """
    #     from sklearn.decomposition import PCA

    #     out_pdf = self.output_prefix + "_cds_rpm_pca_plot.pdf"
    #     # out_png = self.output_prefix + "_rpf_pcaplot.png"

    #     # run the PCA with sklearn function
    #     out_txt = self.output_prefix + "_cds_rpm_pca.txt"

    #     pca = PCA(n_components=2)
    #     pca.fit(np.log2(self.cds_rpm.T + 1))
    #     variance_ratios = pca.explained_variance_ratio_
    #     # covariance = pca.get_covariance() * 100
    #     components = pca.fit_transform(np.log2(self.cds_rpm.T + 1))

    #     pca_df = pd.DataFrame(data=components, columns=['PC1', 'PC2'])
    #     pca_df.index = self.sample_name
    #     pca_df = pca_df.reset_index().rename(columns={'index': 'samples'})

    #     pca_df.to_csv(out_txt, sep='\t', index=True)

    #     pca_df_t = pca_df.reset_index().melt(id_vars="index",
    #                                          var_name="pc",
    #                                          value_name="value").rename(columns={'index': 'samples'})

    #     # draw the PCA figure
    #     fig = px.scatter(pca_df, x="PC1", y="PC2", color="samples")

    #     fig.update_xaxes(tickformat='.2e')
    #     fig.update_yaxes(tickformat='.2e')
    #     fig.update_layout(xaxis_title="PC1 (" + str("{:.2f}".format(variance_ratios[0])) + "%)",
    #                       yaxis_title="PC2 (" + str("{:.2f}".format(variance_ratios[1])) + "%)",
    #                       template="simple_white",
    #                       height=400, width=490)

    #     fig.write_image(out_pdf, engine="kaleido")

    def draw_rpf_heatmap1(self):
        """
        draw the gene rpm heatmap of each samples

        1. log the cds rpm
        2. scale the rpm to center 0
        3. draw the heatmap with di-cluster
        """

        out_pdf = self.output_prefix + "_cds_rpm_heatmap.pdf"
        out_png = self.output_prefix + "_cds_rpm_heatmap.png"

        cds_rpm = self.cds_rpm[(self.cds_rpm > 0).any(axis=1)]
        rpm_log = np.log2(cds_rpm + 1)

        plt.figure(figsize=(9, 12), dpi=300)

        sns.clustermap(rpm_log,
                       cmap="RdBu",
                       center=0,
                       dendrogram_ratio=(.1, .2),
                       cbar_pos=(.02, .02, .025, .17))

        plt.savefig(fname=out_pdf)
        plt.savefig(fname=out_png)

        plt.close()


    def draw_rpf_heatmap2(self):
        """
        Draw the gene RPM heatmap of each sample using matplotlib & seaborn

        1. Log-transform the CDS RPM
        2. Cluster rows using hierarchical clustering (Ward + Euclidean)
        3. Draw clustered heatmap
        """

        out_pdf = self.output_prefix + "_cds_rpm_heatmap.pdf"

        # Step 1: log2 transform (remove all-zero rows)
        rpm_log = np.log2(self.cds_rpm[(self.cds_rpm > 0).any(axis=1)] + 1)

        # Step 2: hierarchical clustering on rows
        cds_dist = pdist(rpm_log, metric='euclidean')
        clustering = hierarchy.linkage(cds_dist, method='ward')
        row_idx = hierarchy.leaves_list(clustering)

        # Reorder matrix according to clustering
        rpm_clustered = rpm_log.iloc[row_idx, :]

        # step 3: row scale
        rpm_scaled = rpm_clustered.sub(rpm_clustered.mean(axis=1), axis=0)
        rpm_scaled = rpm_scaled.div(rpm_scaled.std(axis=1), axis=0)

        # Step 4: draw heatmap
        plt.figure(figsize=(max(6, rpm_clustered.shape[1] * 0.4), 8))
        sns.heatmap(
            rpm_scaled,
            cmap="RdBu_r",
            center=0,
            cbar_kws={'label': 'log2 RPM (scaled)'},
            xticklabels=True,
            yticklabels=True
        )

        plt.title("Expression pattern", fontsize=14, pad=12)
        plt.xticks(rotation=90)
        plt.yticks(rotation=0)
        plt.tight_layout()
        plt.savefig(out_pdf)
        plt.close()
        
    # def draw_rpf_heatmap2(self):
    #     """
    #     draw the gene rpm heatmap of each samples with plotly

    #     1. log the cds rpm
    #     2. calculate the distance of rpm data matrix and clustered with scipy
    #     3. draw the di-cluster heatmap with plotly
    #     """

    #     out_pdf = self.output_prefix + "_cds_rpm_heatmap.pdf"
    #     # out_png = self.output_prefix + "_rpf_heatmap.png"

    #     rpm_log = np.log2(self.cds_rpm[(self.cds_rpm > 0).any(axis=1)] + 1)

    #     cds_dist = pdist(rpm_log)
    #     clustering = hierarchy.linkage(cds_dist, method='ward', metric='euclidean')

    #     row_idx = hierarchy.leaves_list(clustering)
    #     rpm_clustered = rpm_log.iloc[row_idx, :]

    #     fig = go.Figure(
    #         data=go.Heatmap(x=rpm_clustered.columns.values,
    #                         y=rpm_clustered.index.values,
    #                         z=rpm_clustered,
    #                         # zmin=-3, zmax=3,
    #                         colorscale='RdBu',
    #                         reversescale=True)
    #     )

    #     fig.update_layout(xaxis=dict(side='bottom'),
    #                       yaxis=dict(side='left'),
    #                       height=600, width=450,
    #                       title=dict(text='Expression pattern', x=0.5, xanchor='center'),
    #                       xaxis_showgrid=False, yaxis_showgrid=False,
    #                       xaxis_tickangle=-90, yaxis_tickangle=0)

    #     fig.write_image(out_pdf, engine="kaleido")
    #     # fig.write_image(out_png, engine="kaleido")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Project      : RiboParser
@Script       : fq_length.py
@Environment  : python 3.8.5
@Version      : 1.0
@Author       : Rensc 
@Time         : 2025/07/29 20:57:55
@E-mail       : rensc0718@163.com
@License      : (C)Copyright 2023-2025, Ren Shuchao
'''


import argparse
import gzip
import os
from collections import defaultdict

def parse_args():
    parser = argparse.ArgumentParser(description="Count read lengths from FASTQ/FASTQ.GZ files")
    parser.add_argument("-i", "--input", nargs="+", required=True,
                        help="One or more input FASTQ or FASTQ.GZ files")
    parser.add_argument("-o", "--output", default="fq_length_distr.txt",
                        help="Output file name (default: fq_length_distr.txt)")
    return parser.parse_args()

def open_file(filename):
    """Open a file in text mode, using gzip if needed"""
    if filename.endswith(".gz"):
        return gzip.open(filename, "rt")
    else:
        return open(filename, "r")

def get_sample_name(filepath):
    """Extract sample name by removing path and file extension"""
    basename = os.path.basename(filepath)
    for ext in [".fq.gz", ".fastq.gz", ".fq", ".fastq"]:
        if basename.endswith(ext):
            return basename.replace(ext, "")
    return basename

def count_read_lengths(fq_path):
    """Count number of reads by length in a FASTQ file"""
    length_counter = defaultdict(int)
    with open_file(fq_path) as f:
        line_count = 0
        for line in f:
            line_count += 1
            if line_count % 4 == 2:  # sequence line
                read_len = len(line.strip())
                length_counter[read_len] += 1
    return length_counter

def main():
    args = parse_args()
    output_lines = []

    for fq_file in args.input:
        sample_name = get_sample_name(fq_file)
        print(f"Processing sample: {sample_name} ({fq_file})")
        len_counts = count_read_lengths(fq_file)
        for length, count in sorted(len_counts.items()):
            output_lines.append(f"{sample_name}\t{length}\t{count}")

    with open(args.output, "w") as out:
        out.write("Sample\tLength\tRead_Count\n")
        out.write("\n".join(output_lines))
    print(f"\nFinished. Results saved to: {args.output}")

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : rpf_coverage.py


from utils.ribo import ArgsParser
from utils.ribo import Percentage


def main():
    ArgsParser.now_time()
    print('\nDraw the metagene coverage.', flush=True)
    print('\nStep1: Checking the input Arguments.', flush=True)
    args = ArgsParser.percentage_args_parser()
    percent = Percentage.Percentage(args)

    print('\nStep2: Import the RPFs file.', flush=True)
    percent.read_rpf()
    percent.import_gene()

    print('\nStep3: Calculate the percentage of RPFs coverage.', flush=True)
    percent.calc_density_percent()

    print('\nStep4: Draw the histogram and boxplot of RPFs coverage.', flush=True)
    percent.draw_rpf_histogram()
    percent.draw_rpf_boxplot()

    print('\nStep5: Output the RPFs coverage.', flush=True)
    percent.output_density_percent()

    print('\nAll done.\n', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Project      : RiboParser
@Script       : simulate_fastq.py
@Environment  : python 3.8.5
@Version      : 1.0
@Author       : Rensc 
@Time         : 2025/07/29 21:02:35
@E-mail       : rensc0718@163.com
@License      : (C)Copyright 2023-2025, Ren Shuchao
'''


import argparse
import random

def parse_args():
    parser = argparse.ArgumentParser(description="Generate simulated FASTQ reads with random sequence and quality")
    parser.add_argument("-q", "--qmin", type=int, required=True,
                        help="Minimum quality score (e.g. 30)")
    parser.add_argument("-Q", "--qmax", type=int, required=True,
                        help="Maximum quality score (e.g. 45)")
    parser.add_argument("-m", "--minlen", type=int, required=True,
                        help="Minimum read length (e.g. 75)")
    parser.add_argument("-M", "--maxlen", type=int, required=True,
                        help="Maximum read length (e.g. 150)")
    parser.add_argument("-n", "--number", type=int, required=True,
                        help="Number of reads to generate")
    parser.add_argument("-o", "--output", default="simulated.fq",
                        help="Output FASTQ file name (default: simulated.fq)")
    return parser.parse_args()

def random_seq(length):
    """Generate random DNA sequence of given length"""
    return ''.join(random.choices("ACGT", k=length))

def random_qual(length, qmin, qmax):
    """Generate random Phred+33 quality string"""
    return ''.join(chr(random.randint(qmin, qmax) + 33) for _ in range(length))

def main():
    args = parse_args()

    if args.qmin > args.qmax:
        raise ValueError("qmin must be less than or equal to qmax")
    if args.minlen > args.maxlen:
        raise ValueError("minlen must be less than or equal to maxlen")

    with open(args.output, "w") as out:
        for i in range(1, args.number + 1):
            length = random.randint(args.minlen, args.maxlen)
            seq = random_seq(length)
            qual = random_qual(length, args.qmin, args.qmax)
            out.write(f"@read_{i}\n{seq}\n+\n{qual}\n")

    print(f"Generated {args.number} reads in: {args.output}")

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : rpf_Offset_RSBM.py


from utils.ribo import ArgsParser
from utils.ribo.Offset_RSBM import *


def main():
    ArgsParser.now_time()
    print('\nDetect the p-site offset.', flush=True)
    print('\nStep1: Checking the input Arguments.', flush=True)
    args = ArgsParser.offset_rsbm_args_parser()
    offset_attr = Offset(args)

    print('\nStep2: Import the transcripts annotation.', flush=True)
    offset_attr.read_transcript()

    print('\nStep3: Import the bam file.', flush=True)
    offset_attr.get_mrna_reads()

    print('\nStep4: Detect the RSBM offset of sequence profile.', flush=True)
    offset_attr.get_frame_offset()
    offset_attr.format_frame_offset()
    offset_attr.adjust_frame_offset()

    print('\nStep5: Output the frame offset.', flush=True)
    offset_attr.write_frame_offset()

    print('\nStep6: Draw the frame offset heatmap.', flush=True)
    offset_attr.draw_frame_heatmap()

    print('\nAll done.', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Project      : riboParser
@Script       : Periodicity.py
@Environment  : python 3.8.5
@Version      : 1.0
@Author       : Rensc 
@Time         : 2024/01/22 14:13:22
@E-mail       : rensc0718@163.com
@License      : (C)Copyright 2023-2025, Ren Shuchao
'''


from . import RPFs
from collections import OrderedDict
import matplotlib
import matplotlib.pyplot as plt
import polars as pl
import pandas as pd
import numpy as np
import seaborn as sns
import math

class Periodicity(object):

    def __init__(self, args):
        # input and output
        self.transcript = args.transcript
        self.rpf = args.rpf
        self.output = args.output

        # filter the range around tis/tts
        self.rpf_num = args.min
        self.tis = args.tis
        self.tts = args.tts

        # for the high expression gene filter
        self.sample_num = 0
        self.raw_rpf = None
        self.high_gene = None

        self.period = None

    def import_rpf(self):
        '''
        @Message  : function for rpf import.
        @Input    : self.ribo --> rpf density file
                    self.tis --> tis range
                    self.tts --> tts range
                    self.rpf_num --> rpf number filter
        @Return   : output --> rpf density dataframe
        '''

        # raw_rpf, sample_name, sample_num, merged_rpf, total_rpf_num, gene_rpf_sum, high_gene, high_rpf
        rpf_results = RPFs.import_rpf(rpf_file=self.rpf,
                                      sites='P',
                                      frame='all',
                                      sample_num=None,
                                      sample_name=None,
                                      tis = self.tis,
                                      tts = self.tts,
                                      gene=self.transcript,
                                      rpf_num=self.rpf_num)
        
        self.raw_rpf = rpf_results[0].to_pandas()
        # self.sample_name = rpf_results[1]
        self.sample_num = rpf_results[2]
        # self.merged_rpf = rpf_results[3]
        # self.total_rpf_num = rpf_results[4]
        # self.gene_rpf_sum = rpf_results[5]
        self.high_gene = rpf_results[6]
        # self.high_rpf = rpf_results[7]

        del rpf_results

    def calc_3nt_period(self):
        '''
        @Message  : function for 3nt periodicity calculation.
        @Input    : param --> rpf density file
        @Return   : output --> summary of 3nt periodicity
        @Flow     : step1 --> run
        '''

        raw_rpf = self.raw_rpf.drop(columns=["name", "now_nt", "from_tis", "from_tts", "region", "codon"])
        
        period = raw_rpf.apply(np.sum, axis=0).to_frame()
        period = period.reset_index()
        period.columns = ["Name", "Count"]
        period["Sample"] = period['Name'].str[:-3]
        period["Frame"] = period['Name'].str[-1::]
        period["Ratio"] = period["Count"] / period.groupby("Sample")["Count"].transform("sum") * 100

        self.period = period[["Sample", "Frame", "Count", "Ratio"]]

    def output_meta(self):
        out_txt = self.output + "_periodicity.txt"
        self.period.to_csv(out_txt, sep='\t', index=False)

    def get_sub_plot_num(self):
        factor1 = int(self.sample_num ** 0.5)
        factor2 = int(self.sample_num ** 0.5)
        square = factor1 * factor2

        while not square >= self.sample_num:
            if factor2 > factor1:
                factor1 += 1
            else:
                factor2 += 1
            square = factor1 * factor2

        return factor1, factor2
    
    def draw_3nt_period_count(self):

        out_pdf = self.output + "_count_periodicity_plot.pdf"
        out_png = self.output + "_count_periodicity_plot.png"

        nrow, ncol = self.get_sub_plot_num()
        
        matplotlib.use('AGG')
        fig, axes = plt.subplots(nrows=nrow, ncols=ncol, figsize=(nrow * 4, ncol * 3), sharey=True)

        # FIX: ensure axes is always an array
        if not isinstance(axes, np.ndarray):
            axes = np.array([axes])
        axes = axes.flatten()

        for (sample, group), ax in zip(self.period.groupby('Sample'), axes.flatten()):
            bars = group.plot(x='Frame', y='Count', kind='bar', ax=ax, title=sample, legend=False)
            ax.set_ylabel('Count')
            ax.set_xlabel('Frame')

            for bar in bars.patches:
                height = bar.get_height()
                ax.annotate(f'{height:.0f}', 
                            xy=(bar.get_x() + bar.get_width() / 2, height),
                            xytext=(0, 3),  # 3 points vertical offset
                            textcoords="offset points",
                            ha='center', va='bottom')
        
        plt.tight_layout()

        # plt.show()
        fig.savefig(fname=out_pdf)
        fig.savefig(fname=out_png, dpi=300)

    def draw_3nt_period_ratio(self):

        out_pdf = self.output + "_ratio_periodicity_plot.pdf"
        out_png = self.output + "_ratio_periodicity_plot.png"

        nrow, ncol = self.get_sub_plot_num()

        matplotlib.use('AGG')
        fig, axes = plt.subplots(nrows=nrow, ncols=ncol, figsize=(nrow * 4, ncol * 3), sharey=True)

        # FIX: ensure axes is always an array
        if not isinstance(axes, np.ndarray):
            axes = np.array([axes])
        axes = axes.flatten()
        
        for (sample, group), ax in zip(self.period.groupby('Sample'), axes.flatten()):
            bars = group.plot(x='Frame', y='Ratio', kind='bar', ax=ax, title=sample, legend=False)
            ax.set_ylabel('Ratio (%)')
            ax.set_xlabel('Frame')

            for bar in bars.patches:
                height = bar.get_height()
                ax.annotate(f'{height:.0f}', 
                            xy=(bar.get_x() + bar.get_width() / 2, height),
                            xytext=(0, 3),  # 3 points vertical offset
                            textcoords="offset points",
                            ha='center', va='bottom')
        
        plt.tight_layout()

        # plt.show()
        fig.savefig(fname=out_pdf)
        fig.savefig(fname=out_png, dpi=300)
